# coding: utf-8

from __future__ import unicode_literals

import pytest
import pytz
from datetime import datetime

from django.db import models
from django.db import connections
from django.test.utils import CaptureQueriesContext
from freezegun import freeze_time

from zero_downtime_migrations.backend.schema import DatabaseSchemaEditor
from test_app.models import TestModel

pytestmark = pytest.mark.django_db
connection = connections['default']
schema_editor = DatabaseSchemaEditor


def column_classes(model):
    with connection.cursor() as cursor:
        columns = {
            d[0]: (connection.introspection.get_field_type(d[1], d), d)
            for d in connection.introspection.get_table_description(
                cursor,
                model._meta.db_table,
            )
        }
    return columns


def test_sqlmigrate_add_field_working():
    field = models.BooleanField(default=True)
    field.set_attributes_from_name("bool_field")
    with schema_editor(connection=connection, collect_sql=True) as editor:
        editor.add_field(TestModel, field)
        assert editor.collected_sql == [
            "SELECT IS_NULLABLE, DATA_TYPE, COLUMN_DEFAULT from information_schema.columns where table_name = 'test_app_testmodel' and column_name = 'bool_field';",
            'ALTER TABLE "test_app_testmodel" ADD COLUMN "bool_field" boolean NULL;',
            'ALTER TABLE "test_app_testmodel" ALTER COLUMN "bool_field" SET DEFAULT true;',
            "SELECT reltuples::BIGINT FROM pg_class WHERE relname = 'test_app_testmodel';",
            '\n                       WITH cte AS (\n                       SELECT id as pk\n                       FROM test_app_testmodel\n                       WHERE  bool_field is null\n                       LIMIT  1000\n                       )\n                       UPDATE test_app_testmodel table_\n                       SET bool_field = true\n                       FROM   cte\n                       WHERE  table_.id = cte.pk\n                       ;',
            'ALTER TABLE "test_app_testmodel" ALTER COLUMN "bool_field" SET NOT NULL;',
            'ALTER TABLE "test_app_testmodel" ALTER COLUMN "bool_field" DROP DEFAULT;'
        ]


def test_add_bool_field_no_existed_objects_success():
    columns = column_classes(TestModel)
    assert "bool_field" not in columns

    field = models.BooleanField(default=True)
    field.set_attributes_from_name("bool_field")

    with CaptureQueriesContext(connection) as ctx, schema_editor(connection=connection) as editor:
        editor.add_field(TestModel, field)

    columns = column_classes(TestModel)
    assert columns['bool_field'][0] == "BooleanField"
    queries = [query_data['sql'] for query_data in ctx.captured_queries if 'test_app' in query_data['sql']]
    expected_queries = [("SELECT IS_NULLABLE, DATA_TYPE, COLUMN_DEFAULT from information_schema.columns where "
                         "table_name = 'test_app_testmodel' and column_name = 'bool_field';"),
                        'ALTER TABLE "test_app_testmodel" ADD COLUMN "bool_field" boolean NULL',
                        'ALTER TABLE "test_app_testmodel" ALTER COLUMN "bool_field" SET DEFAULT true',
                        "SELECT reltuples::BIGINT FROM pg_class WHERE relname = 'test_app_testmodel';",
                        'SELECT COUNT(*) FROM test_app_testmodel;',
                        'ALTER TABLE "test_app_testmodel" ALTER COLUMN "bool_field" SET NOT NULL',
                        'ALTER TABLE "test_app_testmodel" ALTER COLUMN "bool_field" DROP DEFAULT',
                        ]
    assert queries == expected_queries


def test_add_bool_field_with_existed_object_success(test_object):
    columns = column_classes(TestModel)
    assert "bool_field" not in columns

    field = models.BooleanField(default=True)
    field.set_attributes_from_name("bool_field")
    with CaptureQueriesContext(connection) as ctx, schema_editor(connection=connection) as editor:
        editor.add_field(TestModel, field)
        queries = [query_data['sql'] for query_data in ctx.captured_queries if
                   'test_app' in query_data['sql']]

    columns = column_classes(TestModel)
    assert columns['bool_field'][0] == "BooleanField"
    expected_queries = [("SELECT IS_NULLABLE, DATA_TYPE, COLUMN_DEFAULT from information_schema.columns where "
                         "table_name = 'test_app_testmodel' and column_name = 'bool_field';"),
                        'ALTER TABLE "test_app_testmodel" ADD COLUMN "bool_field" boolean NULL',
                        'ALTER TABLE "test_app_testmodel" ALTER COLUMN "bool_field" SET DEFAULT true',
                        "SELECT reltuples::BIGINT FROM pg_class WHERE relname = 'test_app_testmodel';",
                        'SELECT COUNT(*) FROM test_app_testmodel;',
                        '''
                       WITH cte AS (
                       SELECT id as pk
                       FROM test_app_testmodel
                       WHERE  bool_field is null
                       LIMIT  1000
                       )
                       UPDATE test_app_testmodel table_
                       SET bool_field = true
                       FROM   cte
                       WHERE  table_.id = cte.pk
                       ''',
                        '''
                       WITH cte AS (
                       SELECT id as pk
                       FROM test_app_testmodel
                       WHERE  bool_field is null
                       LIMIT  1000
                       )
                       UPDATE test_app_testmodel table_
                       SET bool_field = true
                       FROM   cte
                       WHERE  table_.id = cte.pk
                       ''',
                        'ALTER TABLE "test_app_testmodel" ALTER COLUMN "bool_field" SET NOT NULL',
                        'ALTER TABLE "test_app_testmodel" ALTER COLUMN "bool_field" DROP DEFAULT',
                        ]
    assert queries == expected_queries
    sql = 'SELECT * from "test_app_testmodel" where id = %s'
    with connection.cursor() as cursor:
        cursor.execute(sql, (test_object.id, ))
        result = cursor.fetchall()
    assert result == [(test_object.id, test_object.name, True)]


def test_add_bool_field_with_existed_many_objects_success(test_object, test_object_two, test_object_three, ):
    columns = column_classes(TestModel)
    assert "bool_field" not in columns

    field = models.BooleanField(default=True)
    field.set_attributes_from_name("bool_field")
    with CaptureQueriesContext(connection) as ctx, schema_editor(connection=connection) as editor:
        editor.add_field(TestModel, field)
        queries = [query_data['sql'] for query_data in ctx.captured_queries if
                   'test_app' in query_data['sql']]

    columns = column_classes(TestModel)
    assert columns['bool_field'][0] == "BooleanField"
    expected_queries = [("SELECT IS_NULLABLE, DATA_TYPE, COLUMN_DEFAULT from information_schema.columns where "
                         "table_name = 'test_app_testmodel' and column_name = 'bool_field';"),
                        'ALTER TABLE "test_app_testmodel" ADD COLUMN "bool_field" boolean NULL',
                        'ALTER TABLE "test_app_testmodel" ALTER COLUMN "bool_field" SET DEFAULT true',
                        "SELECT reltuples::BIGINT FROM pg_class WHERE relname = 'test_app_testmodel';",
                        'SELECT COUNT(*) FROM test_app_testmodel;',
                        '''
                       WITH cte AS (
                       SELECT id as pk
                       FROM test_app_testmodel
                       WHERE  bool_field is null
                       LIMIT  1000
                       )
                       UPDATE test_app_testmodel table_
                       SET bool_field = true
                       FROM   cte
                       WHERE  table_.id = cte.pk
                       ''',
                        '''
                       WITH cte AS (
                       SELECT id as pk
                       FROM test_app_testmodel
                       WHERE  bool_field is null
                       LIMIT  1000
                       )
                       UPDATE test_app_testmodel table_
                       SET bool_field = true
                       FROM   cte
                       WHERE  table_.id = cte.pk
                       ''',
                        'ALTER TABLE "test_app_testmodel" ALTER COLUMN "bool_field" SET NOT NULL',
                        'ALTER TABLE "test_app_testmodel" ALTER COLUMN "bool_field" DROP DEFAULT',
                        ]
    assert queries == expected_queries
    sql = 'SELECT * from "test_app_testmodel" where id = ANY(%s) ORDER BY id'
    with connection.cursor() as cursor:
        cursor.execute(sql, ([test_object.id, test_object_two.id, test_object_three.id], ))
        result = cursor.fetchall()
    assert result == [(test_object.id, test_object.name, True),
                      (test_object_two.id, test_object_two.name, True),
                      (test_object_three.id, test_object_three.name, True),
                      ]


@freeze_time("2017-12-15 03:21:34", tz_offset=-3)
def test_add_datetime_field_no_existed_objects_success():
    columns = column_classes(TestModel)
    assert "datetime_field" not in columns

    field = models.DateTimeField(default=datetime.now)
    field.set_attributes_from_name("datetime_field")

    with CaptureQueriesContext(connection) as ctx, schema_editor(connection=connection) as editor:
        editor.add_field(TestModel, field)
        queries = [query_data['sql'] for query_data in ctx.captured_queries if
                   'test_app' in query_data['sql']]

    columns = column_classes(TestModel)
    assert columns['datetime_field'][0] == "DateTimeField"
    expected_queries = [("SELECT IS_NULLABLE, DATA_TYPE, COLUMN_DEFAULT from information_schema.columns where table_name "
                         "= 'test_app_testmodel' and column_name = 'datetime_field';"),
                        'ALTER TABLE "test_app_testmodel" ADD COLUMN "datetime_field" timestamp with time zone NULL',
                        'ALTER TABLE "test_app_testmodel" ALTER COLUMN "datetime_field" SET DEFAULT \'2017-12-15T00:21:34+00:00\'::timestamptz',
                        "SELECT reltuples::BIGINT FROM pg_class WHERE relname = 'test_app_testmodel';",
                        'SELECT COUNT(*) FROM test_app_testmodel;',
                        'ALTER TABLE "test_app_testmodel" ALTER COLUMN "datetime_field" SET NOT NULL',
                        'ALTER TABLE "test_app_testmodel" ALTER COLUMN "datetime_field" DROP DEFAULT',
                        ]
    assert queries == expected_queries


@freeze_time("2017-12-15 03:21:34", tz_offset=-3)
def test_add_datetime_field_with_existed_object_success(test_object):
    columns = column_classes(TestModel)
    assert "datetime_field" not in columns

    field = models.DateTimeField(default=datetime.now)
    field.set_attributes_from_name("datetime_field")
    with CaptureQueriesContext(connection) as ctx, schema_editor(connection=connection) as editor:
        editor.add_field(TestModel, field)
        queries = [query_data['sql'] for query_data in ctx.captured_queries if
                   'test_app' in query_data['sql']]

    columns = column_classes(TestModel)
    assert columns['datetime_field'][0] == "DateTimeField"
    expected_queries = [("SELECT IS_NULLABLE, DATA_TYPE, COLUMN_DEFAULT from information_schema.columns where "
                         "table_name = 'test_app_testmodel' and column_name = 'datetime_field';"),
                        'ALTER TABLE "test_app_testmodel" ADD COLUMN "datetime_field" timestamp with time zone NULL',
                        'ALTER TABLE "test_app_testmodel" ALTER COLUMN "datetime_field" SET DEFAULT \'2017-12-15T00:21:34+00:00\'::timestamptz',
                        "SELECT reltuples::BIGINT FROM pg_class WHERE relname = 'test_app_testmodel';",
                        'SELECT COUNT(*) FROM test_app_testmodel;',
                        '''
                       WITH cte AS (
                       SELECT id as pk
                       FROM test_app_testmodel
                       WHERE  datetime_field is null
                       LIMIT  1000
                       )
                       UPDATE test_app_testmodel table_
                       SET datetime_field = \'2017-12-15T00:21:34+00:00\'::timestamptz
                       FROM   cte
                       WHERE  table_.id = cte.pk
                       ''',
                        '''
                       WITH cte AS (
                       SELECT id as pk
                       FROM test_app_testmodel
                       WHERE  datetime_field is null
                       LIMIT  1000
                       )
                       UPDATE test_app_testmodel table_
                       SET datetime_field = \'2017-12-15T00:21:34+00:00\'::timestamptz
                       FROM   cte
                       WHERE  table_.id = cte.pk
                       ''',
                        'ALTER TABLE "test_app_testmodel" ALTER COLUMN "datetime_field" SET NOT NULL',
                        'ALTER TABLE "test_app_testmodel" ALTER COLUMN "datetime_field" DROP DEFAULT',
                        ]
    assert queries == expected_queries
    sql = 'SELECT * from "test_app_testmodel" where id = %s'
    with connection.cursor() as cursor:
        cursor.execute(sql, (test_object.id, ))
        result = cursor.fetchall()
    assert result == [(test_object.id, test_object.name, datetime(2017, 12, 15, 0, 21, 34, tzinfo=pytz.UTC))]


@freeze_time("2017-12-15 03:21:34", tz_offset=-3)
def test_add_datetime_field_with_existed_many_objects_success(test_object, test_object_two, test_object_three, ):
    columns = column_classes(TestModel)
    assert "datetime_field" not in columns

    field = models.DateTimeField(default=datetime.now)
    field.set_attributes_from_name("datetime_field")
    with CaptureQueriesContext(connection) as ctx, schema_editor(connection=connection) as editor:
        editor.add_field(TestModel, field)
        queries = [query_data['sql'] for query_data in ctx.captured_queries if
                   'test_app' in query_data['sql']]

    columns = column_classes(TestModel)
    assert columns['datetime_field'][0] == "DateTimeField"
    expected_queries = [("SELECT IS_NULLABLE, DATA_TYPE, COLUMN_DEFAULT from information_schema.columns where "
                         "table_name = 'test_app_testmodel' and column_name = 'datetime_field';"),
                        'ALTER TABLE "test_app_testmodel" ADD COLUMN "datetime_field" timestamp with time zone NULL',
                        'ALTER TABLE "test_app_testmodel" ALTER COLUMN "datetime_field" SET DEFAULT \'2017-12-15T00:21:34+00:00\'::timestamptz',
                        "SELECT reltuples::BIGINT FROM pg_class WHERE relname = 'test_app_testmodel';",
                        'SELECT COUNT(*) FROM test_app_testmodel;',
                        '''
                       WITH cte AS (
                       SELECT id as pk
                       FROM test_app_testmodel
                       WHERE  datetime_field is null
                       LIMIT  1000
                       )
                       UPDATE test_app_testmodel table_
                       SET datetime_field = \'2017-12-15T00:21:34+00:00\'::timestamptz
                       FROM   cte
                       WHERE  table_.id = cte.pk
                       ''',
                        '''
                       WITH cte AS (
                       SELECT id as pk
                       FROM test_app_testmodel
                       WHERE  datetime_field is null
                       LIMIT  1000
                       )
                       UPDATE test_app_testmodel table_
                       SET datetime_field = \'2017-12-15T00:21:34+00:00\'::timestamptz
                       FROM   cte
                       WHERE  table_.id = cte.pk
                       ''',
                        'ALTER TABLE "test_app_testmodel" ALTER COLUMN "datetime_field" SET NOT NULL',
                        'ALTER TABLE "test_app_testmodel" ALTER COLUMN "datetime_field" DROP DEFAULT',
                        ]
    assert queries == expected_queries
    sql = 'SELECT * from "test_app_testmodel" where id = ANY(%s) ORDER BY id'
    with connection.cursor() as cursor:
        cursor.execute(sql, ([test_object.id, test_object_two.id, test_object_three.id], ))
        result = cursor.fetchall()
    assert result == [(test_object.id, test_object.name, datetime(2017, 12, 15, 0, 21, 34, tzinfo=pytz.UTC)),
                      (test_object_two.id, test_object_two.name, datetime(2017, 12, 15, 0, 21, 34, tzinfo=pytz.UTC)),
                      (test_object_three.id, test_object_three.name, datetime(2017, 12, 15, 0, 21, 34, tzinfo=pytz.UTC)),
                      ]




# coding: utf-8

from __future__ import unicode_literals

import sys
import inspect

from distutils.version import StrictVersion

try:
    from django.db.backends.postgresql.schema import DatabaseSchemaEditor as BaseEditor
except ImportError:
    from django.db.backends.postgresql_psycopg2.schema import DatabaseSchemaEditor as BaseEditor

import django
from django.db.models.fields import NOT_PROVIDED
from django.db.models.fields.related import RelatedField
from django.db import transaction
from django.db.migrations.questioner import InteractiveMigrationQuestioner

DJANGO_VERISON = StrictVersion(django.get_version())


class ZeroDownTimeMixin(object):
    sql_estimate_count_in_table = "SELECT reltuples::BIGINT FROM pg_class WHERE relname = '%(table)s';"
    sql_count_in_table = "SELECT COUNT(*) FROM %(table)s;"
    sql_count_in_table_with_null = "SELECT COUNT(*) FROM %(table)s WHERE %(column)s is NULL;"
    sql_update_batch = '''
                       WITH cte AS (
                       SELECT %(pk_column_name)s as pk
                       FROM %(table)s
                       WHERE  %(column)s is null
                       LIMIT  %(batch_size)s
                       )
                       UPDATE %(table)s table_
                       SET %(column)s = %(value)s
                       FROM   cte
                       WHERE  table_.%(pk_column_name)s = cte.pk
                       '''
    sql_check_column_status = ("SELECT IS_NULLABLE, DATA_TYPE, COLUMN_DEFAULT from information_schema.columns "
                               "where table_name = '%(table)s' and column_name = '%(column)s';")

    def add_field(self, model, field):
        if isinstance(field, RelatedField) or field.default is NOT_PROVIDED:
            return super(ZeroDownTimeMixin, self).add_field(model, field)
        else:
            # Checking which actions we should perform - maybe this operation was run
            # before and it had crashed for some reason
            actions = self.get_actions_to_perform(model, field)
            if len(actions) == 0:
                return

            # Saving initial values
            default_effective_value = self.effective_default(field)
            nullable = field.null
            # Update the values to the required ones
            field.default = None if DJANGO_VERISON < StrictVersion('1.11') else NOT_PROVIDED
            if nullable is False:
                field.null = True

            # For Django < 1.10
            atomic = getattr(self, 'atomic_migration', True)

            if self.connection.in_atomic_block:
                self.atomic.__exit__(None, None, None)

            available_args = {
                'model': model,
                'field': field,
                'nullable': nullable,
                'default_effective_value': default_effective_value,
            }
            # Performing needed actions
            for action in actions:
                action = '_'.join(action.split())
                func = getattr(self, action)
                func_args = {arg: available_args[arg] for arg in
                             inspect.getargspec(func).args if arg != 'self'
                             }
                func(**func_args)

            # If migrations was atomic=True initially
            # entering atomic block again
            if atomic:
                self.atomic = transaction.atomic(self.connection.alias)
                self.atomic.__enter__()


    def add_field_with_default(self, model, field, default_effective_value):
        """
        Adding field with default in two separate
        operations, so we can avoid rewriting the
        whole table
        """
        with transaction.atomic():
            super(ZeroDownTimeMixin, self).add_field(model, field)
            self.add_default(model, field, default_effective_value)

    def update_existing_rows(self, model, field, default_effective_value):
        """
        Updating existing rows in table by (relatively) small batches
        to avoid long locks on table
        """
        objects_in_table = self.count_objects_in_table(model=model)
        if objects_in_table > 0:
            objects_in_batch_count = self.get_objects_in_batch_count(objects_in_table)
            while True:
                with transaction.atomic():
                    updated = self.update_batch(model=model, field=field,
                                                objects_in_batch_count=objects_in_batch_count,
                                                value=default_effective_value,
                                                )
                    print('Update {} rows in {}'.format(updated, model._meta.db_table))
                    if updated is None or updated == 0:
                        break

    def set_not_null_for_field(self, model, field, nullable):
        # If field was not null - adding
        # this knowledge to table
        if nullable is False:
            self.set_not_null(model, field)

    def get_column_info(self, model, field):
        sql = self.sql_check_column_status % {
            "table": model._meta.db_table,
            "column": field.name,
        }
        return self.get_query_result(sql)

    def get_actions_to_perform(self, model, field):
        actions = [
            'add field with default',
            'update existing rows',
            'set not null for field',
            'drop default',
        ]

        # Checking maybe this column already exists
        # if so asking user what to do next
        column_info = self.get_column_info(model, field)

        if column_info is not None:
            existed_nullable, existed_type, existed_default = column_info

            questioner = InteractiveMigrationQuestioner()
            question_template = ('It look like column "{}" in table "{}" already exist with following '
                                 'parameters: TYPE: "{}", DEFAULT: "{}", NULLABLE: "{}".'
                                 )
            question = question_template.format(field.name, model._meta.db_table,
                                                existed_type, existed_default,
                                                existed_nullable,
                                                )
            choices = ('abort migration',
                       'drop column and run migration from beginning',
                       'manually choose action to start from',
                       'show how many rows still need to be updated',
                       'mark operation as successful and proceed to next operation',
                       )

            result = questioner._choice_input(question, choices)
            if result == 1:
                sys.exit(1)
            elif result == 2:
                self.remove_field(model, field)
            elif result == 3:
                question = 'Now choose from which action process should continue'
                result = questioner._choice_input(question, actions)
                actions = actions[result-1:]
            elif result == 4:
                question = 'Rows in table where column is null: "{}"'
                need_to_update = self.need_to_update(model=model, field=field)
                questioner._choice_input(question.format(need_to_update),
                                         ('Continue', )
                                         )
                return self.get_actions_to_perform(model, field)
            elif result == 5:
                actions = []
        return actions

    def get_pk_column_name(self, model):
        return model._meta.pk.name

    def update_batch(self, model, field, objects_in_batch_count, value):
        pk_column_name = self.get_pk_column_name(model)
        sql = self.sql_update_batch % {
            "table": model._meta.db_table,
            "column": field.name,
            "batch_size": objects_in_batch_count,
            "pk_column_name": pk_column_name,
            "value": "%s",
        }
        params = [value]
        return self.get_query_result(sql, params, row_count=True)

    def get_objects_in_batch_count(self, model_count):
        """
        Calculate batch size

        :param model_count: int
        :return: int
        """
        if model_count > 500000:
            value = 10000
        else:
            value = int((model_count / 100) * 5)
        return max(1000, value)

    def get_query_result(self, sql, params=(), row_count=False):
        """
        Default django backend execute function does not
        return any result so we use this custom where needed
        """
        if self.collect_sql:
            # in collect_sql case use django function logic
            return self.execute(sql, params)

        with self.connection.cursor() as cursor:
            cursor.execute(sql, params)
            if row_count:
                return cursor.rowcount
            return cursor.fetchone()

    def parse_cursor_result(self, cursor, place=0, collect_sql_value=1,):
        if self.collect_sql:
            result = collect_sql_value  # For sqlmigrate purpose
        else:
            result = cursor[place]
        return result

    def execute_table_query(self, sql, model):
        sql = sql % {
            "table": model._meta.db_table
        }
        cursor = self.get_query_result(sql)
        return self.parse_cursor_result(cursor=cursor)

    def count_objects_in_table(self, model):
        count = self.execute_table_query(sql=self.sql_estimate_count_in_table,
                                         model=model,
                                         )
        if count == 0:
            # Check, maybe statistic is outdated?
            # Because previous count return 0 it will be fast query
            count = self.execute_table_query(sql=self.sql_count_in_table,
                                             model=model,
                                             )
        return count

    def need_to_update(self, model, field):
        sql = self.sql_count_in_table_with_null % {
            "table": model._meta.db_table,
            "column": field.name,
        }
        cursor = self.get_query_result(sql)
        return self.parse_cursor_result(cursor=cursor)

    def drop_default(self, model, field):
        set_default_sql, params = self._alter_column_default_sql(field, drop=True)
        self.execute_alter_column(model, set_default_sql, params)

    def add_default(self, model, field, default_value):
        set_default_sql, params = self._alter_column_default_sql(field, default_value)
        self.execute_alter_column(model, set_default_sql, params)

    def set_not_null(self, model, field):
        set_not_null_sql = self.generate_set_not_null(field)
        self.execute_alter_column(model, set_not_null_sql)

    def execute_alter_column(self, model, changes_sql, params=()):
        sql = self.sql_alter_column % {
            "table": self.quote_name(model._meta.db_table),
            "changes": changes_sql,
        }
        self.execute(sql, params)

    def generate_set_not_null(self, field):
        new_db_params = field.db_parameters(connection=self.connection)
        sql = self.sql_alter_column_not_null
        return sql % {
                'column': self.quote_name(field.column),
                'type': new_db_params['type'],
        }

    def _alter_column_default_sql(self, field, default_value=None, drop=False):
        """
        Copy this method from django2.0
        https://github.com/django/django/blob/master/django/db/backends/base/schema.py#L787
        """
        default = '%s'
        params = [default_value]

        if drop:
            params = []

        new_db_params = field.db_parameters(connection=self.connection)
        sql = self.sql_alter_column_no_default if drop else self.sql_alter_column_default
        return (
            sql % {
                'column': self.quote_name(field.column),
                'type': new_db_params['type'],
                'default': default,
            },
            params,
        )

    def execute(self, sql, params=()):
        exit_atomic = False
        # Account for non-string statement objects.
        sql = str(sql)

        if 'CREATE INDEX' in sql:
            exit_atomic = True
            sql = sql.replace('CREATE INDEX', 'CREATE INDEX CONCURRENTLY')
        atomic = self.connection.in_atomic_block
        if exit_atomic and atomic:
            self.atomic.__exit__(None, None, None)

        super(ZeroDownTimeMixin, self).execute(sql, params)

        if exit_atomic and atomic:
            self.atomic = transaction.atomic(self.connection.alias)
            self.atomic.__enter__()


class DatabaseSchemaEditor(ZeroDownTimeMixin, BaseEditor):
    pass



# coding: utf-8

from __future__ import unicode_literals

import pytest
import re

from django.db import models
from django.db import connections
from django.test.utils import CaptureQueriesContext

from zero_downtime_migrations.backend.schema import DatabaseSchemaEditor
from test_app.models import TestModel

connection = connections['default']
schema_editor = DatabaseSchemaEditor


@pytest.mark.django_db(transaction=True)
def test_create_index_success():
    old_field = models.IntegerField()
    old_field.set_attributes_from_name("name")

    field = models.IntegerField(db_index=True)
    field.set_attributes_from_name("name")
    pattern = r'CREATE INDEX CONCURRENTLY "test_app_testmodel_name_\w+(_uniq)?" ON "test_app_testmodel" \("name"\)'
    with CaptureQueriesContext(connection) as ctx, schema_editor(connection=connection) as editor:
        editor.alter_field(TestModel, old_field, field)
        assert len(ctx.captured_queries) == 1
        assert re.search(pattern, ctx.captured_queries[0]['sql']) is not None


def test_sqlmigrate_create_index_working():
    old_field = models.IntegerField()
    old_field.set_attributes_from_name("name")

    field = models.IntegerField(db_index=True)
    field.set_attributes_from_name("name")
    pattern = r'CREATE INDEX CONCURRENTLY "test_app_testmodel_name_\w+(_uniq)?" ON "test_app_testmodel" \("name"\)'
    with schema_editor(connection=connection, collect_sql=True) as editor:
        editor.alter_field(TestModel, old_field, field)
        assert len(editor.collected_sql) == 1
        assert re.search(pattern, editor.collected_sql[0]) is not None
